"""
Command line interface for rysk_client
"""

import os
import sys

import rich_click as click

from rysk_client.client import RyskClient
from rysk_client.src.constants import NULL_ADDRESS
from rysk_client.src.position_side import PositionSide
from rysk_client.src.utils import get_logger, render_table


def set_logger(ctx, level):
    """Set the logger."""
    if not hasattr(ctx, "logger"):
        ctx.logger = get_logger()
        ctx.logger.setLevel(level)
    return ctx.logger


def set_client(ctx):
    """Set the client."""
    if not hasattr(ctx, "client"):
        auth = {
            "address": os.environ["ETH_ADDRESS"],
            "private_key": os.environ["ETH_PRIVATE_KEY"],
            "logger": ctx.logger,
        }
        ctx.client = RyskClient(**auth)
        if not ctx.client.web3_client.web3.isConnected():
            raise ConnectionError("Web3 client not connected.")
    return ctx.client


@click.group("Rysk client")
@click.option(
    "--log-level",
    "-l",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    help="Logging level.",
)
@click.pass_context
def cli(ctx, log_level):
    """Rysk client command line interface."""
    ctx.ensure_object(dict)
    ctx.obj["logger"] = set_logger(ctx, log_level)
    ctx.obj["client"] = set_client(ctx)


@cli.group("markets")
def markets():
    """Interact with markets."""


@cli.group("tickers")
def tickers():
    """Interact with tickers."""


@cli.group("positions")
def positions():
    """Interact with positions."""


@cli.group("trades")
def trades():
    """Interact with trades."""


@cli.group("balances")
@click.pass_context
def balances(ctx):  # pylint: disable=unused-argument
    """Interact with balances."""


@balances.command("fetch")
@click.pass_context
def fetch_balances(ctx):
    """Fetch balances."""
    client = ctx.obj["client"]
    columns = [
        "symbol",
        "balance",
    ]
    current_balances = [
        {"symbol": k, "balance": v} for k, v in client.fetch_balances().items()
    ]

    render_table("Balances", current_balances, columns)


@markets.command("fetch")
@click.option("--sort", "-s", default="expiration", help="Sort by column.")
@click.option("--is_active", "-a", default=True, help="Filter by active.")
@click.pass_context
def fetch_markets(ctx, sort, is_active):
    """Fetch markets."""
    client = ctx.obj["client"]
    markets = client.fetch_markets()
    if is_active:
        markets = [market for market in markets if market["active"]]

    columns = [
        "id",
        "expiration",
        "strike",
        "bid",
        "ask",
        "delta",
    ]
    sorted_markets = sorted(markets, key=lambda x: int(x[sort]))
    render_table("Markets", sorted_markets, columns)


@tickers.command("fetch")
@click.option("--sort", "-s", default="expiration", help="Sort by column.")
@click.option("--market", "-m", default=None, help="Filter by market.")
@click.pass_context
def fetch_tickers(ctx, sort, market):
    """Fetch tickers."""
    client = ctx.obj["client"]
    tickers = client.fetch_tickers(market)
    sorted_tickers = sorted(tickers, key=lambda x: int(x[sort]))
    columns = ["id", "expiration", "bid", "ask"]
    render_table("Tickers", sorted_tickers, columns)


@positions.command("list")
@click.option("--expired", "-e", is_flag=True, help="Include expired positions.")
@click.option("--market", "-m", default=None, help="Filter by market.")
@click.option("--is_open", "-o", is_flag=True, help="Filter by open positions")
@click.pass_context
def list_positions(ctx, expired, market, is_open):
    """List positions retriueved fromt the lens contract."""
    client = ctx.obj["client"]
    logger = ctx.obj["logger"]
    logger.info(
        f"Fetching positions for {client._crypto.address}"  # pylint: disable=protected-access
    )
    positions = client.fetch_positions(expired=expired)
    columns = [
        "symbol",
        "side",
        "entryPrice",
        "id",
        "size",
        "unrealizedPnl",
        "realizedPnl",
    ]
    if market:
        positions = [p for p in positions if p["symbol"] == market]
    if is_open:
        positions = [p for p in positions if p["size"] != 0]

    render_table("Positions", positions, columns)


@trades.command("watch")
@click.pass_context
def watch(ctx):
    """Watch trades as they occur on the contract."""
    client = ctx.obj["client"]
    client.watch_trades()


@positions.command("close")
@click.pass_context
@click.option("--market", "-m", required=True, help="Market to close.")
@click.option(
    "--size",
    "-s",
    required=False,
    help="Size to close. Default is All.",
    default=None,
    type=click.FLOAT,
)
def close(ctx, market, size):
    """Close a position."""
    client = ctx.obj["client"]
    logger = ctx.obj["logger"]
    user_address = client._crypto.address  # pylint: disable=protected-access
    positions = client.fetch_positions()
    if market not in (f["symbol"] for f in positions):
        raise ValueError(
            f"User {user_address} does not have an open position in {market}"
        )
    positions = [p for p in positions if p["symbol"] == market]
    position = [f for f in positions if f["size"] != 0]
    if len(position) > 1:
        raise ValueError(
            f"User {user_address} has multiple open positions in {market}. Please visit the Rysk UI to close positions."
        )
    if len(position) == 0:
        raise ValueError(
            f"User {user_address} does not have an open position in {market}"
        )
    position = position.pop()
    logger.info(f"Closing position for {user_address}")

    if position["side"] == PositionSide.LONG.value:
        txn = client.close_long(market, size)
    else:
        txn = client.close_short(market, size)

    if txn:
        logger.info(f"Transaction hash: {txn}")
    else:
        logger.info("No transaction hash returned. Check logs for errors.")


@positions.command("settle")
@click.option(
    "--vault_ids",
    "-v",
    required=True,
    type=click.STRING,
)
@click.pass_context
def settle(ctx, vault_ids):
    """Settle a position."""
    logger = ctx.obj["logger"]
    client = ctx.obj["client"]
    vault_ids_list = [int(vault_id) for vault_id in vault_ids.split(",")]
    user_address = client._crypto.address  # pylint: disable=protected-access

    logger.info(f"Settling vault {vault_ids} for {user_address}")
    if not client.web3_client.web3.is_connected():
        raise ConnectionError("Web3 client not connected.")
    vaults = dict(client.web3_client.fetch_user_vaults(user_address))
    for vault_id in vault_ids_list:

        if vault_id not in vaults:
            raise ValueError(f"Vault {vault_id} not found for {user_address}")
        if vaults[vault_id] == NULL_ADDRESS:
            raise ValueError(
                f"Vault {vault_id} has already been settled for {user_address}"
            )
    for vault_id in vault_ids_list:
        result = client.settle_vault(int(vault_id))
        if result:
            logger.info(f"Successfully settled vault {vault_id}. Transaction: {result}")
        else:
            logger.error(f"Failed to settle vault {vault_id}.")


# we pass a list of markets to redeem
@positions.command("redeem")
@click.option(
    "--markets",
    "-m",
    required=True,
    type=click.STRING,
    help="Comma separated list of markets to redeem. For example: ETH-02JUN23-1900-P, ETH-02JUN23-2000-P",
)
@click.pass_context
def redeem(ctx, markets):
    """Redeem a position."""
    logger = ctx.obj["logger"]
    client = ctx.obj["client"]
    markets_list = markets.split(",")
    user_address = client._crypto.address  # pylint: disable=protected-access
    logger.info(f"Redeeming {markets} for {user_address}")
    for market in markets_list:
        result = client.redeem_market(market)
        if result:
            logger.info(f"Successfully redeemed {market}. Transaction: {result}")
        else:
            logger.error(f"Failed to redeem {market}.")
            raise ValueError(f"Failed to redeem {market}.")


@positions.command("collateralize")
def collateralize():
    """Collateralize a position."""
    raise NotImplementedError


@trades.command("list")
def list_trades():
    """List trades."""
    raise NotImplementedError  # disable=raising-to-general-error


@trades.command("create")
@click.option("--market", "-m", required=True, help="Market to trade.")
@click.option("--side", "-s", required=True, help="Buy or sell.")
@click.option(
    "--amount", "-a", required=True, type=click.FLOAT, help="Size of the trade."
)
@click.pass_context
def create_trade(ctx, market, side, amount):
    """Create a trade."""
    client = ctx.obj["client"]
    logger = ctx.obj["logger"]
    trade = client.create_order(market, amount, side)
    if not trade:
        logger.error("Failed to create trade.")
        sys.exit(1)
    logger.info(f"Created trade: {trade}")


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""
Rysk Client Operation Manager
"""


from dataclasses import dataclass
from enum import Enum

from rysk_client.src.action_type import ActionType, RyskActionType
from rysk_client.src.collateral import CollateralFactory
from rysk_client.src.constants import (EMPTY_SERIES, NULL_ADDRESS, NULL_DATA,
                                       Chain)
from rysk_client.src.rysk_option_market import MarketFactory, RyskOptionMarket
from rysk_client.src.utils import from_wei_to_opyn


class OperationType(Enum):
    """Distinguish between operations for the rysk and opyn contracts."""

    OPYN_ACTION = 0
    RYSK_ACTION = 1


@dataclass
class OperationFactory:
    """Create operations for the rysk and opyn contracts."""

    chain: Chain

    def __init__(self, chain: Chain):
        self.chain = chain
        self.collateral_factory = CollateralFactory(chain)
        self.market_factory = MarketFactory(chain)

    def buy(
        self,
        acceptable_premium: int,
        owner_address: str,
        amount: int,
        option_market: RyskOptionMarket,
        issuance_required: bool = False,
    ):
        """Create the operation to buy an option."""
        operations = []

        if issuance_required:
            operations.append(
                {
                    "actionType": RyskActionType.ISSUE.value,
                    "owner": NULL_ADDRESS,
                    "secondAddress": NULL_ADDRESS,
                    "asset": NULL_ADDRESS,
                    "vaultId": 0,
                    "amount": 0,
                    "optionSeries": self.market_factory.to_series(option_market),
                    "indexOrAcceptablePremium": 0,
                    "data": NULL_ADDRESS,
                }
            )

        operations.append(
            {
                "actionType": RyskActionType.BUY_OPTION.value,
                "owner": NULL_ADDRESS,
                "secondAddress": owner_address,
                "asset": NULL_ADDRESS,
                "vaultId": 0,
                "amount": amount,
                "optionSeries": self.market_factory.to_series(option_market),
                "indexOrAcceptablePremium": acceptable_premium,
                "data": NULL_ADDRESS,
            }
        )
        return [
            {
                "operation": OperationType.RYSK_ACTION.value,
                "operationQueue": operations,
            }
        ]

    def close_long(
        self,
        acceptable_premium: int,
        owner_address: str,
        otoken_address: str,
        amount: int,
    ):
        """Create the operation to buy an option."""
        return [
            {
                "operation": OperationType.RYSK_ACTION.value,
                "operationQueue": [
                    {
                        "actionType": RyskActionType.SELL_OPTION.value,
                        "owner": NULL_ADDRESS,
                        "secondAddress": owner_address,
                        "asset": otoken_address,
                        "vaultId": 0,
                        "amount": amount,
                        "optionSeries": EMPTY_SERIES,
                        "indexOrAcceptablePremium": acceptable_premium,
                        "data": NULL_DATA,
                    }
                ],
            }
        ]

    def sell(
        self,
        acceptable_premium: int,
        owner_address: str,
        exchange_address: str,
        otoken_address: str,
        amount: int,
        vault_id: int,
        collateral_amount: int,
        rysk_option_market: RyskOptionMarket,
        issue_new_vault: bool = False,
    ):
        """Create the operation to sell an option."""
        if rysk_option_market.is_put:
            collateral = self.collateral_factory.USDC

        else:
            collateral = self.collateral_factory.WETH

        required_data = [
            {
                "actionType": ActionType.DEPOSIT_COLLATERAL.value,
                "owner": owner_address,
                "secondAddress": exchange_address,
                "asset": collateral,
                "vaultId": vault_id,
                "amount": collateral_amount,
                "optionSeries": EMPTY_SERIES,
                "indexOrAcceptablePremium": 0,
                "data": NULL_DATA,
            },
            {
                "actionType": ActionType.MINT_SHORT_OPTION.value,
                "owner": owner_address,
                "secondAddress": exchange_address,
                "asset": otoken_address,
                "vaultId": vault_id,
                "amount": from_wei_to_opyn(amount),
                "optionSeries": EMPTY_SERIES,
                "indexOrAcceptablePremium": 0,
                "data": NULL_DATA,
            },
        ]
        if issue_new_vault:
            # we need to open a vault
            required_data = [
                {
                    "actionType": ActionType.OPEN_VAULT.value,
                    "owner": owner_address,
                    "secondAddress": owner_address,
                    "asset": NULL_ADDRESS,
                    "vaultId": vault_id,
                    "amount": 0,
                    "optionSeries": EMPTY_SERIES,
                    "indexOrAcceptablePremium": 0,
                    "data": NULL_DATA,
                }
            ] + required_data
        return [
            {
                "operation": OperationType.OPYN_ACTION.value,
                "operationQueue": required_data,
            },
            {
                "operation": OperationType.RYSK_ACTION.value,
                "operationQueue": [
                    {
                        "actionType": RyskActionType.SELL_OPTION.value,
                        "owner": NULL_ADDRESS,
                        "secondAddress": owner_address,
                        "asset": NULL_ADDRESS,
                        "vaultId": 0,
                        "amount": amount,
                        "optionSeries": {
                            "expiration": int(rysk_option_market.expiration),
                            "strike": int(rysk_option_market.strike),
                            "isPut": rysk_option_market.is_put,
                            "underlying": self.collateral_factory.WETH,
                            "strikeAsset": self.collateral_factory.USDC,
                            "collateral": self.collateral_factory.WETH
                            if not rysk_option_market.is_put
                            else self.collateral_factory.USDC,
                        },
                        "indexOrAcceptablePremium": int(acceptable_premium),
                        "data": NULL_DATA,
                    }
                ],
            },
        ]

    def close_short(
        self,
        acceptable_premium: int,
        owner_address: str,
        otoken_address: str,
        amount: int,
        collateral_amount: int,
        collateral_asset: str,
        vault_id: int,
    ):
        """
        Create the operation to close a short options
        """

        tx_data = [
            {
                "operation": OperationType.RYSK_ACTION.value,
                "operationQueue": [
                    {
                        "actionType": RyskActionType.BUY_OPTION.value,
                        "owner": NULL_ADDRESS,
                        "secondAddress": owner_address,
                        "asset": otoken_address,
                        "vaultId": 0,
                        "amount": amount,
                        "optionSeries": EMPTY_SERIES,
                        "indexOrAcceptablePremium": acceptable_premium,
                        "data": NULL_DATA,
                    }
                ],
            },
            {
                "operation": OperationType.OPYN_ACTION.value,
                "operationQueue": [
                    {
                        "actionType": ActionType.BURN_SHORT_OPTION.value,
                        "owner": owner_address,
                        "secondAddress": owner_address,
                        "asset": otoken_address,
                        "vaultId": vault_id,
                        "amount": from_wei_to_opyn(amount),
                        "optionSeries": EMPTY_SERIES,
                        "indexOrAcceptablePremium": 0,
                        "data": NULL_DATA,
                    },
                    {
                        "actionType": ActionType.WITHDRAW_COLLATERAL.value,
                        "owner": owner_address,
                        "secondAddress": owner_address,
                        "asset": collateral_asset,
                        "vaultId": vault_id,
                        "amount": collateral_amount,
                        "optionSeries": EMPTY_SERIES,
                        "indexOrAcceptablePremium": 0,
                        "data": NULL_DATA,
                    },
                ],
            },
        ]
        return tx_data



"""
PnlCalculator class is used to calculate the PnL of a position.
"""


class Trade:
    """Class to represent a trade."""

    def __init__(
        self,
        quantity=None,
        price=None,
        total_cost=None,
    ):
        if sum([quantity is None, price is None, total_cost is None]) != 1:
            raise ValueError(
                "2 out of 3 of quantity, price, total_cost must be provided"
            )

        if quantity is not None and price is not None:
            self._total_cost = quantity * price
            self._quantity = quantity
            self._price = price
        elif quantity is not None and total_cost is not None:
            self._price = total_cost / quantity
            self._quantity, self._total_cost = quantity, total_cost

        elif price is not None and total_cost is not None:
            self._quantity = total_cost / price
            self._price, self._total_cost = price, total_cost

    @property
    def quantity(self):
        """Return the quantity of the trade."""
        return self._quantity

    @property
    def price(self):
        """Return the price of the trade."""
        return self._price

    @property
    def total_cost(self):
        """Return the total cost of the trade."""
        return self._total_cost

    def __str__(self) -> str:
        """Format the trade as a string."""
        return f"Trade(quantity={self.quantity}, price={self.price})"

    def __repr__(self) -> str:
        """Format the trade as a string."""
        return str(self)


class PnlCalculator:
    """Class to calculate the PnL of a position."""

    def __init__(self):
        """A PnlCalculator is initialised with no trades."""
        self.current_price = 0
        self.position_size = 0
        self.total_cost = 0
        self.realised_pnl_total = 0

    def update_price(self, price):
        """Update the current price of the PnlCalculator."""
        self.current_price = price

    @property
    def average_price(self):
        """Calculate the average price of the open position."""
        if self.position_size == 0:
            return 0

        return self.total_cost / self.position_size

    @property
    def unrealised_pnl(self):
        """Calculate the unrealised PnL of the PnlCalculator."""
        return self.position_size * (self.current_price - self.average_price)

    @property
    def realised_pnl(self):
        """Return the realised PnL of the PnlCalculator."""
        return self.realised_pnl_total

    def add_trades(self, trades):
        """Add a list of trades to the PnlCalculator."""
        for trade in trades:
            self.add_trade(trade)

    def add_trade(self, trade):
        """
        add a trade to the PnlCalculator.
        positions can be short
        trades can be added in any order

        buy_trade = Trade(1, 1000)
        sell_trade = Trade(-1, 1010)
        pnl_calculator.add_trade(buy_trade)
        pnl_calculator.add_trade(sell_trade)
        assert pnl_calculator.position_size == 0
        assert pnl_calculator.realised_pnl == 10

        """
        # handle open
        if self.position_size == 0:
            self.position_size = trade.quantity
            self.total_cost = trade.total_cost
            return

        # handle partial close
        if self.position_size * trade.quantity < 0:
            # we are closing part of the position
            # we need to calculate the realised pnl
            # and update the position size and total cost
            realised_pnl = -self.position_size * (trade.price - self.average_price)
            self.realised_pnl_total -= realised_pnl
            self.position_size += trade.quantity
            self.total_cost += trade.total_cost
            return

        # handle full close
        if self.position_size + trade.quantity == 0:
            # we are closing the position
            # we need to calculate the realised pnl
            # and update the position size and total cost
            realised_pnl = -self.position_size * (trade.price - self.average_price)
            self.realised_pnl_total -= realised_pnl
            self.position_size = 0
            self.total_cost = 0
            return

        # handle increase position
        if self.position_size * trade.quantity > 0:
            # we are increasing the position
            # we need to update the position size and total cost
            self.position_size += trade.quantity
            self.total_cost += trade.total_cost
            return



"""
This module contains the RyskOptionMarket class.
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from rysk_client.src.collateral import CollateralFactory
from rysk_client.src.constants import USDC_MULTIPLIER, WETH_MULTIPLIER, Chain

EXPIRATION_TIME = "08:00:00"


@dataclass
class TradingSpec:
    """
    Class to represent the trading spec.
    """

    iv: int  # noqa: C0103
    quote: int
    fee: int
    disabled: bool
    premium_too_small: bool


@dataclass
class OptionStrikeDrill:
    """
    Class to represent the option strike drill.
    """

    strike: int
    sell: TradingSpec
    buy: TradingSpec
    delta: int
    exposure: int
    series_collateral_exchange_balance_usdc: Optional[int] = None
    series_collateral_exchange_balance_weth: Optional[int] = None


class OptionChain:
    """Class to represent the option chain."""

    _raw_data: tuple
    expirations: List[int]
    strikes: Dict[int, List[int]]

    def __init__(self, raw_data: tuple) -> None:
        self._raw_data = raw_data
        self.parse_expirations()
        self.parse_strikes()

    def parse_expirations(self):
        """Parse and set the expirations."""
        self.expirations = self._raw_data[0]

    def parse_strikes(self):
        """Parse and set the strikes."""
        self.strikes = {}
        for index, expiration_1 in enumerate(self.expirations):
            expiration_data = self._raw_data[1][index]
            (
                expiration_3,  # noqa: F841
                call_strikes,  # noqa: F841
                call_data,
                put_strikes,  # noqa: F841
                put_data,
                underlying_value,  # noqa: F841
            ) = expiration_data

            call_option_drill = self._parse_option_data(call_data)
            put_option_drill = self._parse_option_data(put_data)

            self.strikes[expiration_1] = {
                "call": call_option_drill,
                "put": put_option_drill,
            }

    def _parse_option_data(self, option_data: tuple):
        """
        Parse the option data.
        """
        markets = []
        if len(option_data[0]) == 5:
            cols = [
                "strike",
                "sell_trading_specs",
                "buy_trading_specs",
                "delta",
                "exposure",
            ]
        elif len(option_data[0]) == 7:
            cols = [
                "strike",
                "sell_trading_specs",
                "buy_trading_specs",
                "delta",
                "exposure",
                "series_collateral_exchange_balance_usdc",
                "series_collateral_exchange_balance_weth",
            ]
        else:
            raise ValueError("Option data has unexpected length. is not beta or prod.")

        for row in option_data:
            params = {}
            for k, val in zip(cols, row):
                if k in ["sell_trading_specs", "buy_trading_specs"]:
                    params[k.split("_")[0]] = TradingSpec(*val)
                else:
                    params[k] = val
            markets.append(OptionStrikeDrill(**params))
        return markets

    @property
    def active_markets(self):
        """Return the active strikes for the active markets, where option drill is not disabled."""
        active_markets = []
        for expiration, option_drill in self.strikes.items():
            for option_type, option_type_drill in option_drill.items():
                for option_market in option_type_drill:
                    if (
                        not option_market.sell.disabled
                        or not option_market.buy.disabled
                    ):
                        active_markets.append((expiration, option_type, option_market))
        return active_markets

    @property
    def current_price(self):
        """
        Return the underlying price from the oracle.
        """
        return self._raw_data[1][-1][-1] / WETH_MULTIPLIER


class OptionType(Enum):
    """Option type enum."""

    CALL = "call"
    PUT = "put"


@dataclass
class RyskOptionMarket:  # pylint: disable=too-many-instance-attributes
    """Rysk option market."""

    strike: float
    expiration: int
    is_put: bool
    active: bool = True
    # market data
    bid: Optional[int] = None
    ask: Optional[int] = None
    dhv: Optional[int] = None
    delta: Optional[int] = None
    _collateral: Optional[str] = None

    @classmethod
    def from_series(cls, series):
        """Returns a RyskOptionMarket from a series"""
        return cls(
            strike=series["strike"],
            expiration=series["expiration"],
            is_put=series["isPut"],
        )

    def _parse_name(self):
        """Returns the name of the option market into the following format.

        NAME = "ETH-30JUN23-2200-C"
        """
        _type = "P" if self.is_put else "C"
        _expiration = datetime.fromtimestamp(self.expiration).strftime("%d%b%y").upper()
        _strike = int(self.strike / WETH_MULTIPLIER)
        return f"ETH-{_expiration}-{_strike}-{_type}"

    @property
    def name(self):
        """Returns the name of the option market"""
        return self._parse_name()

    def __str__(self):
        return f"RyskOptionMarket({self.name})"

    @classmethod
    def from_option_drill(
        cls, expiration: int, option_type: str, option_drill: OptionStrikeDrill
    ):
        """Returns a RyskOptionMarket from an option drill"""
        return cls(
            strike=option_drill.strike,
            expiration=expiration,
            is_put=option_type == "put",
            ask=option_drill.buy.quote,
            bid=option_drill.sell.quote,
            dhv=option_drill.exposure,
            delta=option_drill.delta,
        )

    def to_json(self):
        """Returns the option market as a json"""
        market_data = {}
        if self.bid and self.ask and self.dhv is not None:
            market_data = {
                "bid": self.bid / USDC_MULTIPLIER,
                "ask": self.ask / USDC_MULTIPLIER,
                "dhv": self.dhv / WETH_MULTIPLIER,
            }
        result = {
            "id": self.name,
            "strike": self.strike / WETH_MULTIPLIER,
            "expiration": self.expiration,
            "optionType": "put" if self.is_put else "call",
            "active": self.active,
        }
        if self.delta:
            result["delta"] = self.delta / WETH_MULTIPLIER
        result.update(**market_data)
        return result

    @classmethod
    def from_json(cls, json):
        """Returns a RyskOptionMarket from a json"""
        return cls(
            strike=json["strike"] * WETH_MULTIPLIER,
            expiration=json["expiration"],
            is_put=json["optionType"] == "put",
            active=json["active"],
            bid=json.get("bid") * USDC_MULTIPLIER,
            ask=json.get("ask") * USDC_MULTIPLIER,
            dhv=json.get("dhv") * WETH_MULTIPLIER,
            delta=json.get("delta"),
        )

    @classmethod
    def from_str(cls, name: str):
        """Returns a RyskOptionMarket from a name"""
        _name = name.split("-")
        expiration_date = datetime.strptime(
            f"{_name[1]}T{EXPIRATION_TIME}+00:00",
            "%d%b%yT%H:%M:%S%z",
        )
        _expiration = int(expiration_date.timestamp())

        _strike = int(_name[2]) * WETH_MULTIPLIER
        _is_put = _name[3] == "P"
        return cls(_strike, _expiration, _is_put)

    @property
    def collateral(self):
        """Returns the collateral of the option market"""
        if self._collateral is None:
            raise ValueError("Collateral not set")
        return self._collateral

    @collateral.setter
    def collateral(self, collateral):
        """Sets the collateral of the option market"""
        self._collateral = collateral


@dataclass
class MarketFactory:
    """Class to generate option markets."""

    chain: Chain

    def __init__(self, chain: Chain, covered: bool = True):
        """
        Markets factory will generate option markets for the given chain.
        If covered is True, it will generate markets for the covered options.
        I.e. for a given strike and expiration, it will generate a call and a put.
        If called with covered=False, it will generate markets for the naked options.
        """
        self.chain = chain
        self.collateral_factory = CollateralFactory(chain)
        self._covered = covered

    def to_series(self, rysk_option_market: RyskOptionMarket):
        """Returns the series of the option market
        effectively, we are adding the appriate contract addresses to the option market.
        """
        return {
            "strike": int(rysk_option_market.strike),
            "expiration": int(rysk_option_market.expiration),
            "isPut": rysk_option_market.is_put,
            "underlying": self.collateral_factory.WETH,
            "strikeAsset": self.collateral_factory.USDC,
            "collateral": rysk_option_market.collateral,
        }



"""
Helper functions for the rysk_client package.
"""
import json
import logging
import sys
from copy import deepcopy
from dataclasses import asdict
from typing import Any, Dict, List, Optional

from rich import print_json
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table
from web3 import Web3

from rysk_client.src.constants import (ARBITRUM_GOERLI, DEFAULT_ENCODING,
                                       PROTOCOL_DEPLOYMENTS)


def from_wei_to_opyn(amount: int):
    """Convert amount from wei to opyn."""
    return int(amount / 10**10)


def get_logger():
    """Get the logger."""
    logger = logging.getLogger(__name__)
    formatter = logging.Formatter("%(message)s")

    # we check if the logger already has a handler
    # to avoid adding multiple handlers
    if logger.hasHandlers():
        return logger
    if sys.stdout.isatty():

        handler = RichHandler(
            markup=False,
            rich_tracebacks=True,
            locals_max_string=None,
            locals_max_length=None,
        )
    else:
        handler = logging.StreamHandler(sys.stdout)

    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger


def get_contract(name, web3, chain, address=None):
    """Returns a web3 contract instance for the given contract name"""
    spec = PROTOCOL_DEPLOYMENTS[chain.name].contracts[name]

    with open(spec.path, "r", encoding=DEFAULT_ENCODING) as abi:
        abi = json.loads(abi.read())["abi"]
    res = asdict(spec)
    del res["path"]
    res["abi"] = abi
    if address is not None:
        res["address"] = address
    return web3.eth.contract(**res)


def get_web3(chain=ARBITRUM_GOERLI) -> Web3:
    """Returns a web3 instance connected to RPC_URL"""

    web3 = Web3(Web3.HTTPProvider(chain.rpc_url))
    return web3


def render_table(title: str, data, cols: Optional[List[str]] = None):
    """Render a table from a dynamic input."""
    table = Table(title=title)
    # we first create the columns

    if cols is None:
        columns = []
        for row in data:
            for column in row.keys():
                if column not in columns:
                    columns.append(column)
    else:
        columns = cols

    for column in columns:
        table.add_column(column)
    # then we add the rows
    for row in data:
        table.add_row(*[str(row.get(column, "")) for column in columns])
    console = Console()
    console.print(table)


def print_operate_tuple(operate_tuple: List[Dict[str, Any]]):
    """
    Ensure that the operate tuple is formated ina manner compatible with
    tenderly's api.
    print it using rich.
    """
    display_tuple = deepcopy(operate_tuple)
    keys_to_stringify = [
        "strike",
        "amount",
        "indexOrAcceptablePremium",
        "vaultId",
        "actionType",
    ]

    def stringify_json(json_data: Any):
        """
        Recursively stringify json data.
        """
        if isinstance(json_data, dict):
            for key, value in json_data.items():
                if key in keys_to_stringify:
                    json_data[key] = str(value)
                else:
                    stringify_json(value)
        elif isinstance(json_data, list):
            for value in json_data:
                stringify_json(value)

    stringify_json(display_tuple)
    print_json(data=display_tuple)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import Address, LedgerApi

from rysk_client.packages.eightballer.contracts.opyn_controller import \
    PUBLIC_ID


class OpynController(Contract):  # pylint: disable=too-many-public-methods
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def addressbook(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'addressbook' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.addressbook().call()
        return {"address": result}

    @classmethod
    def calculator(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'calculator' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.calculator().call()
        return {"address": result}

    @classmethod
    def call_restricted(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'call_restricted' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.callRestricted().call()
        return {"bool": result}

    @classmethod
    def can_settle_assets(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _underlying: Address,
        _strike: Address,
        _collateral: Address,
        _expiry: int,
    ) -> JSONLike:
        """Handler method for the 'can_settle_assets' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.canSettleAssets(
            _underlying=_underlying,
            _strike=_strike,
            _collateral=_collateral,
            _expiry=_expiry,
        ).call()
        return {"bool": result}

    @classmethod
    def full_pauser(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'full_pauser' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.fullPauser().call()
        return {"address": result}

    @classmethod
    def get_account_vault_counter(
        cls, ledger_api: LedgerApi, contract_address: str, _account_owner: Address
    ) -> JSONLike:
        """Handler method for the 'get_account_vault_counter' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getAccountVaultCounter(
            _accountOwner=_account_owner
        ).call()
        return {"int": result}

    @classmethod
    def get_configuration(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'get_configuration' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getConfiguration().call()
        return {
            "address_0": result[0],
            "address_1": result[1],
            "address_2": result[2],
            "address_3": result[3],
        }

    @classmethod
    def get_naked_cap(
        cls, ledger_api: LedgerApi, contract_address: str, _asset: Address
    ) -> JSONLike:
        """Handler method for the 'get_naked_cap' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getNakedCap(_asset=_asset).call()
        return {"int": result}

    @classmethod
    def get_naked_pool_balance(
        cls, ledger_api: LedgerApi, contract_address: str, _asset: Address
    ) -> JSONLike:
        """Handler method for the 'get_naked_pool_balance' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getNakedPoolBalance(_asset=_asset).call()
        return {"int": result}

    @classmethod
    def get_payout(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _otoken: Address,
        _amount: int,
    ) -> JSONLike:
        """Handler method for the 'get_payout' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getPayout(_otoken=_otoken, _amount=_amount).call()
        return {"int": result}

    @classmethod
    def get_proceed(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _vault_id: int,
    ) -> JSONLike:
        """Handler method for the 'get_proceed' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getProceed(
            _owner=_owner, _vault_id=_vault_id
        ).call()
        return {"int": result}

    @classmethod
    def get_vault(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _vault_id: int,
    ) -> JSONLike:
        """Handler method for the 'get_vault' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getVault(_owner=_owner, _vault_id=_vault_id).call()
        return {"tuple": result}

    @classmethod
    def get_vault_liquidation_details(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _vault_id: int,
    ) -> JSONLike:
        """Handler method for the 'get_vault_liquidation_details' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getVaultLiquidationDetails(
            _owner=_owner, _vault_id=_vault_id
        ).call()
        return {"address": result, "int_0": result, "int_1": result}

    @classmethod
    def get_vault_with_details(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _vault_id: int,
    ) -> JSONLike:
        """Handler method for the 'get_vault_with_details' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getVaultWithDetails(
            _owner=_owner, _vault_id=_vault_id
        ).call()
        return {"tuple": result[0], "int_0": result[1], "int_1": result[2]}

    @classmethod
    def has_expired(
        cls, ledger_api: LedgerApi, contract_address: str, _otoken: Address
    ) -> JSONLike:
        """Handler method for the 'has_expired' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.hasExpired(_otoken=_otoken).call()
        return {"bool": result}

    @classmethod
    def is_liquidatable(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _vault_id: int,
    ) -> JSONLike:
        """Handler method for the 'is_liquidatable' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.isLiquidatable(
            _owner=_owner, _vault_id=_vault_id
        ).call()
        return {"bool": result, "int_1": result, "int_2": result}

    @classmethod
    def is_operator(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _owner: Address,
        _operator: Address,
    ) -> JSONLike:
        """Handler method for the 'is_operator' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.isOperator(
            _owner=_owner, _operator=_operator
        ).call()
        return {"bool": result}

    @classmethod
    def is_settlement_allowed(
        cls, ledger_api: LedgerApi, contract_address: str, _otoken: Address
    ) -> JSONLike:
        """Handler method for the 'is_settlement_allowed' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.isSettlementAllowed(_otoken=_otoken).call()
        return {"bool": result}

    @classmethod
    def oracle(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'oracle' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.oracle().call()
        return {"address": result}

    @classmethod
    def owner(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'owner' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.owner().call()
        return {"address": result}

    @classmethod
    def partial_pauser(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'partial_pauser' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.partialPauser().call()
        return {"address": result}

    @classmethod
    def pool(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'pool' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.pool().call()
        return {"address": result}

    @classmethod
    def system_fully_paused(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'system_fully_paused' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.systemFullyPaused().call()
        return {"bool": result}

    @classmethod
    def system_partially_paused(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'system_partially_paused' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.systemPartiallyPaused().call()
        return {"bool": result}

    @classmethod
    def whitelist(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'whitelist' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.whitelist().call()
        return {"address": result}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import Address, LedgerApi

from rysk_client.packages.eightballer.contracts.opyn_option_registry import \
    PUBLIC_ID


class OpynOptionRegistry(Contract):  # pylint: disable=too-many-public-methods
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def address_book(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'address_book' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.addressBook().call()
        return {"address": result}

    @classmethod
    def authority(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'authority' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.authority().call()
        return {"address": result}

    @classmethod
    def call_lower_health_factor(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'call_lower_health_factor' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.callLowerHealthFactor().call()
        return {"int": result}

    @classmethod
    def call_upper_health_factor(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'call_upper_health_factor' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.callUpperHealthFactor().call()
        return {"int": result}

    @classmethod
    def check_vault_health(
        cls, ledger_api: LedgerApi, contract_address: str, vault_id: int
    ) -> JSONLike:
        """Handler method for the 'check_vault_health' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.checkVaultHealth(vault_id=vault_id).call()
        return {
            "isBelowMin": result,
            "isAboveMax": result,
            "healthFactor": result,
            "upperHealthFactor": result,
            "collatRequired": result,
            "collatAsset": result,
        }

    @classmethod
    def collateral_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'collateral_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.collateralAsset().call()
        return {"address": result}

    @classmethod
    def get_collateral(
        cls, ledger_api: LedgerApi, contract_address: str, series: tuple, amount: int
    ) -> JSONLike:
        """Handler method for the 'get_collateral' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getCollateral(series=series, amount=amount).call()
        return {"int": result}

    @classmethod
    def get_issuance_hash(
        cls, ledger_api: LedgerApi, contract_address: str, _series: tuple
    ) -> JSONLike:
        """Handler method for the 'get_issuance_hash' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getIssuanceHash(_series=_series).call()
        return {"str": result}

    @classmethod
    def get_otoken(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        underlying: Address,
        strike_asset: Address,
        expiration: int,
        is_put: bool,
        strike: int,
        collateral: Address,
    ) -> JSONLike:
        """Handler method for the 'get_otoken' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getOtoken(
            underlying=underlying,
            strike_asset=strike_asset,
            expiration=expiration,
            is_put=is_put,
            strike=strike,
            collateral=collateral,
        ).call()
        return {"address": result}

    @classmethod
    def get_series(
        cls, ledger_api: LedgerApi, contract_address: str, _series: tuple
    ) -> JSONLike:
        """Handler method for the 'get_series' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getSeries(_series=_series).call()
        return {"address": result}

    @classmethod
    def get_series_address(
        cls, ledger_api: LedgerApi, contract_address: str, issuance_hash: str
    ) -> JSONLike:
        """Handler method for the 'get_series_address' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getSeriesAddress(issuance_hash=issuance_hash).call()
        return {"address": result}

    @classmethod
    def get_series_info(
        cls, ledger_api: LedgerApi, contract_address: str, series: Address
    ) -> JSONLike:
        """Handler method for the 'get_series_info' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getSeriesInfo(series=series).call()
        return {"tuple": result}

    @classmethod
    def keeper(
        cls, ledger_api: LedgerApi, contract_address: str, var_0: Address
    ) -> JSONLike:
        """Handler method for the 'keeper' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.keeper(var_0).call()
        return {"bool": result}

    @classmethod
    def liquidity_pool(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'liquidity_pool' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.liquidityPool().call()
        return {"address": result}

    @classmethod
    def put_lower_health_factor(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'put_lower_health_factor' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.putLowerHealthFactor().call()
        return {"int": result}

    @classmethod
    def put_upper_health_factor(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'put_upper_health_factor' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.putUpperHealthFactor().call()
        return {"int": result}

    @classmethod
    def series_info(
        cls, ledger_api: LedgerApi, contract_address: str, var_0: Address
    ) -> JSONLike:
        """Handler method for the 'series_info' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.seriesInfo(var_0).call()
        return {
            "expiration": result,
            "strike": result,
            "is_put": result,
            "underlying": result,
            "strike_asset": result,
            "collateral": result,
        }

    @classmethod
    def vault_count(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'vault_count' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.vaultCount().call()
        return {"int": result}

    @classmethod
    def vault_ids(
        cls, ledger_api: LedgerApi, contract_address: str, var_0: Address
    ) -> JSONLike:
        """Handler method for the 'vault_ids' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.vault_ids(var_0).call()
        return {"int": result}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import Address, LedgerApi

from rysk_client.packages.eightballer.contracts.option_exchange import \
    PUBLIC_ID


class OptionExchange(Contract):  # pylint: disable=too-many-public-methods
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def addressbook(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'addressbook' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.addressbook().call()
        return {"address": result}

    @classmethod
    def approved_collateral(
        cls, ledger_api: LedgerApi, contract_address: str, var_0: Address, var_1: bool
    ) -> JSONLike:
        """Handler method for the 'approved_collateral' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.approvedCollateral(var_0, var_1).call()
        return {"bool": result}

    @classmethod
    def authority(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'authority' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.authority().call()
        return {"address": result}

    @classmethod
    def catalogue(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'catalogue' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.catalogue().call()
        return {"address": result}

    @classmethod
    def check_hash(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        option_series: tuple,
        strike_decimal_converted: int,
        is_sell: bool,
    ) -> JSONLike:
        """Handler method for the 'check_hash' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.checkHash(
            optionSeries=option_series,
            strikeDecimalConverted=strike_decimal_converted,
            isSell=is_sell,
        ).call()
        return {"oHash": result}

    @classmethod
    def collateral_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'collateral_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.collateralAsset().call()
        return {"address": result}

    @classmethod
    def fee_recipient(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'fee_recipient' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.feeRecipient().call()
        return {"address": result}

    @classmethod
    def get_delta(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'get_delta' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getDelta().call()
        return {"delta": result}

    @classmethod
    def get_option_details(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        series_address: Address,
        option_series: tuple,
    ) -> JSONLike:
        """Handler method for the 'get_option_details' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getOptionDetails(
            seriesAddress=series_address, optionSeries=option_series
        ).call()
        return {"address": result, "tuple": result, "int": result}

    @classmethod
    def get_pool_denominated_value(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'get_pool_denominated_value' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getPoolDenominatedValue().call()
        return {"int": result}

    @classmethod
    def held_tokens(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        var_0: Address,
        var_1: Address,
    ) -> JSONLike:
        """Handler method for the 'held_tokens' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.heldTokens(var_0, var_1).call()
        return {"int": result}

    @classmethod
    def liquidity_pool(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'liquidity_pool' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.liquidityPool().call()
        return {"address": result}

    @classmethod
    def max_trade_size(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'max_trade_size' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.maxTradeSize().call()
        return {"int": result}

    @classmethod
    def min_trade_size(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'min_trade_size' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.minTradeSize().call()
        return {"int": result}

    @classmethod
    def paused(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'paused' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.paused().call()
        return {"bool": result}

    @classmethod
    def pool_fees(
        cls, ledger_api: LedgerApi, contract_address: str, var_0: Address
    ) -> JSONLike:
        """Handler method for the 'pool_fees' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.poolFees(var_0).call()
        return {"int": result}

    @classmethod
    def pricer(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'pricer' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.pricer().call()
        return {"address": result}

    @classmethod
    def protocol(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'protocol' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.protocol().call()
        return {"address": result}

    @classmethod
    def strike_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'strike_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.strikeAsset().call()
        return {"address": result}

    @classmethod
    def swap_router(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'swap_router' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.swapRouter().call()
        return {"address": result}

    @classmethod
    def underlying_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'underlying_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.underlyingAsset().call()
        return {"address": result}

    @classmethod
    def update(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'update' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.update().call()
        return {"int": result}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi

from rysk_client.packages.eightballer.contracts.d_h_v_lens import PUBLIC_ID


class DHVLens(Contract):
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def catalogue(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'catalogue' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.catalogue().call()
        return {"address": result}

    @classmethod
    def collateral_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'collateral_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.collateralAsset().call()
        return {"address": result}

    @classmethod
    def exchange(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'exchange' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.exchange().call()
        return {"address": result}

    @classmethod
    def get_expirations(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'get_expirations' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getExpirations().call()
        return {"list[int]": result}

    @classmethod
    def get_option_chain(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'get_option_chain' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getOptionChain().call()
        return {"tuple": result}

    @classmethod
    def get_option_expiration_drill(
        cls, ledger_api: LedgerApi, contract_address: str, expiration: int
    ) -> JSONLike:
        """Handler method for the 'get_option_expiration_drill' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getOptionExpirationDrill(
            expiration=expiration
        ).call()
        return {"tuple": result}

    @classmethod
    def pricer(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'pricer' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.pricer().call()
        return {"address": result}

    @classmethod
    def protocol(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'protocol' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.protocol().call()
        return {"address": result}

    @classmethod
    def strike_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'strike_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.strikeAsset().call()
        return {"address": result}

    @classmethod
    def underlying_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'underlying_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.underlyingAsset().call()
        return {"address": result}



"""
Module to manage positions.
"""
from dataclasses import dataclass
from typing import List, Optional

from rysk_client.src.order import Order
from rysk_client.src.position_side import PositionSide
from rysk_client.src.rysk_option_market import RyskOptionMarket


@dataclass
class Position:
    """Class to represent a position."""

    option_market: RyskOptionMarket
    orders: List[Order]
    side: PositionSide
    collateral: str
    size: float = 0
    pnl: float = 0
    owner_address: Optional[str] = None



"""
This module contains the Order class, which is used to represent an order.
"""
from dataclasses import dataclass

from rysk_client.src.order_side import OrderSide


@dataclass
class Order:
    """Class to represent an order"""

    price: float
    amount: float
    order_id: str
    order_side: OrderSide
    order_type: str = "market"

    def __str__(self) -> str:
        return f"Order({self.order_side}, {self.amount}, {self.price})"

    def __repr__(self) -> str:
        return self.__str__()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import Address, LedgerApi

from rysk_client.packages.eightballer.contracts.weth import PUBLIC_ID


class Weth(Contract):
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def domain_separator(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'domain_separator' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.DOMAIN_SEPARATOR().call()
        return {"str": result}

    @classmethod
    def allowance(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        owner: Address,
        spender: Address,
    ) -> JSONLike:
        """Handler method for the 'allowance' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.allowance(owner=owner, spender=spender).call()
        return {"int": result}

    @classmethod
    def balance_of(
        cls, ledger_api: LedgerApi, contract_address: str, account: Address
    ) -> JSONLike:
        """Handler method for the 'balance_of' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.balanceOf(account=account).call()
        return {"int": result}

    @classmethod
    def decimals(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'decimals' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.decimals().call()
        return {"int": result}

    @classmethod
    def l1_address(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'l1_address' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.l1Address().call()
        return {"address": result}

    @classmethod
    def l2_gateway(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'l2_gateway' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.l2Gateway().call()
        return {"address": result}

    @classmethod
    def name(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'name' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.name().call()
        return {"str": result}

    @classmethod
    def nonces(
        cls, ledger_api: LedgerApi, contract_address: str, owner: Address
    ) -> JSONLike:
        """Handler method for the 'nonces' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.nonces(owner=owner).call()
        return {"int": result}

    @classmethod
    def symbol(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'symbol' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.symbol().call()
        return {"str": result}

    @classmethod
    def total_supply(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'total_supply' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.totalSupply().call()
        return {"int": result}



"""
Web3 client for RyskFinance.
"""
import json
import logging
from collections import deque
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

import websocket
from web3 import HTTPProvider
from web3.contract import Contract

from rysk_client.src.action_type import ActionType
from rysk_client.src.collateral import CollateralFactory
from rysk_client.src.constants import (ARBITRUM_GOERLI, NULL_ADDRESS,
                                       NULL_DATA, Chain)
from rysk_client.src.crypto import EthCrypto
from rysk_client.src.operation_factory import OperationFactory
from rysk_client.src.order import Order
from rysk_client.src.order_side import OrderSide
from rysk_client.src.utils import (get_contract, get_logger, get_web3,
                                   print_operate_tuple)


class Balances(Enum):
    """
    Class to represent the balances of an address.
    """


class Web3Client:  # pylint: disable=too-many-instance-attributes
    """Client for the RyskFinance protocol."""

    beyond_pricer: Contract
    opyn_controller: Contract
    option_exchange: Contract
    option_registry: Contract

    # lenses
    dhv_lens_mk1: Contract
    user_position_lens_mk1: Contract

    # collateral
    usdc: Contract
    weth: Contract

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        crypto: Optional[EthCrypto] = None,
        chain: Chain = ARBITRUM_GOERLI,
        verbose: bool = True,
    ):
        """
        Initialize the client with the web3 provider.
        """
        self.chain = chain
        self.web3: HTTPProvider = get_web3(chain)
        self.beyond_pricer = get_contract("beyond_pricer", self.web3, self.chain)
        self.opyn_controller = get_contract("opyn_controller", self.web3, self.chain)
        self.option_exchange = get_contract("option_exchange", self.web3, self.chain)
        self.option_registry = get_contract(
            "opyn_option_registry", self.web3, self.chain
        )

        self.user_position_lens_mk1 = get_contract(
            "user_position_lens", self.web3, self.chain
        )
        self.dhv_lens_mk1 = get_contract("d_h_v_lens", self.web3, self.chain)

        self.usdc = get_contract("usdc", self.web3, self.chain)
        self.weth = get_contract("weth", self.web3, self.chain)
        # is this an artifact of the testnet?
        self.settlement_usdc = get_contract("usdc", self.web3, self.chain)
        self.settlement_weth = get_contract("weth", self.web3, self.chain)

        self._logger = logger or get_logger()
        self._processed_tx: deque = deque(maxlen=100)
        self._crypto = crypto
        self._verbose = verbose
        self._operation_factory = OperationFactory(self.chain)
        self._collateral_factory = CollateralFactory(self.chain)

    def get_otoken_contract(self, otoken_address: str) -> Contract:
        """
        Get the otoken contract.
        """
        return get_contract(
            "o_token", self.web3, chain=self.chain, address=otoken_address
        )

    def get_options_prices(
        self,
        option_data,
        dhv_exposure,
        amount=1000000000000000000,
        side="buy",
        collateral="weth",
        strike_asset="usdc",
    ):
        """,
        We call the beyond pricer to determine the prices for a market
        huge thanks to 0xPawel2 and Jib &&
        """
        if side not in ["buy", "sell"]:
            raise ValueError("Side must be buy or sell")

        if not self._collateral_factory.is_supported(collateral):
            raise TypeError(f"Collateral {collateral} is not supported")
        # here we call the contract functions

        option_series = (
            int(option_data["expiration"]),
            int(option_data["strike"]),
            bool(option_data["isPut"]),
            self._collateral_factory.from_symbol(collateral),
            self._collateral_factory.from_symbol(strike_asset),
            self._collateral_factory.from_symbol(strike_asset),
        )

        try:
            result = self.beyond_pricer.functions.quoteOptionPrice(
                option_series,
                int(amount),
                side == "sell",
                int(dhv_exposure),
            ).call()
        except Exception as error:  # pylint: disable=broad-except
            self._logger.error(
                f"Error calling beyond pricer: {error} with {option_series}"
            )

            return 0
        return result[0]

    def get_balances(self):
        """
        Get the balances for an address
        """
        safe_address = self.web3.toChecksumAddress(self._crypto.address)
        return {
            "weth": self.settlement_weth.functions.balanceOf(safe_address).call()
            / 1e18,
            "usdc": self.settlement_usdc.functions.balanceOf(safe_address).call() / 1e6,
            "eth": self.web3.eth.get_balance(self._crypto.address) / 1e18,
        }

    def watch_trades(self):
        """
        Subscribe to boeht pending trades and completed trades.
        """
        ws_connection = websocket.WebSocketApp(
            self.chain.wss_url,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        ws_connection.run_forever()

    def on_message(self, websocket, message):  # pylint: disable=unused-argument
        """On new message, process the message."""
        data = json.loads(message)
        if set(data.keys()) == {"id", "result", "jsonrpc"}:
            return

        tx_hash = data["params"]["result"]["transactionHash"]
        if tx_hash in self._processed_tx:
            return
        self._logger.info(f"New Transaction hash: {tx_hash}")
        tx_receipt = self.web3.eth.get_transaction_receipt(tx_hash)
        self._logger.info(f"Sender: {tx_receipt['from']}")
        tx_args, error = self._get_tx_args(tx_receipt)

        if not error:
            order = self.parse_args(tx_args, tx_receipt)
            self._logger.info(order)

        self._processed_tx.append(tx_hash)

    def parse_args(self, args, tx_hash):
        """
        parse the args from the options exchange into an order.
        """
        return Order(
            order_id=tx_hash,
            price=args["premium"] / 1e10,
            amount=args["optionAmount"] / 1e18,
            order_side=OrderSide("buy") if "buyer" in args else OrderSide("sell"),
        )

    def on_error(self, websocket, error):  # pylint: disable=unused-argument
        """On error from the websocket."""
        self._logger.error(error)

    def on_close(self, websocket, *args):  # pylint: disable=unused-argument
        """On closing the connection to the websocket."""
        self._logger.info(f"Connection closed. {args}")

    def on_open(self, websocket):
        """On opening the connection to the websocket."""
        self._logger.info(
            f"Opening connection to Option Exchange at {self.option_exchange.address}"
        )
        subscription_msg_template = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "eth_subscribe",
            "params": ["logs", {"address": self.option_exchange.address}],
        }
        websocket.send(json.dumps(subscription_msg_template))

    def _get_tx_args(self, tx_receipt: str) -> Tuple[Dict, bool]:
        """Get the transaction arguments."""
        try:
            rich_logs = self.option_exchange.events.OptionsSold().processReceipt(tx_receipt)  # type: ignore
            return dict(rich_logs[0]["args"]), False

        except Exception:  # pylint: disable=W0718
            rich_logs = self.option_exchange.events.OptionsBought().processReceipt(tx_receipt)  # type: ignore
            return dict(rich_logs[0]["args"]), False

    def is_approved(
        self, contract: Contract, spender: str, owner: str, amount: int
    ) -> bool:
        """
        Check if the spender is approved to spend the owner's tokens.
        """
        allowance = contract.functions.allowance(owner, spender).call()
        return allowance >= amount

    def create_approval(
        self, contract: Contract, spender: str, owner: str, amount: int
    ) -> dict:
        """
        Create an approval transaction.
        """
        transaction = contract.functions.approve(
            spender, int(amount)
        ).build_transaction(
            {
                **{
                    "from": owner,
                    "nonce": self.web3.eth.get_transaction_count(owner),
                },
                **self._default_tx_params,
            }
        )
        return transaction

    @property
    def _default_tx_params(self) -> dict:
        """
        Default transaction parameters.
        """
        return {
            "gasPrice": int(self.web3.eth.gas_price * 1.5),
        }

    def fetch_user_vaults(self, address: str) -> List[Dict[str, Any]]:  # noqa: W0613
        """
        Fetch the user's positions from the options exchange.
        """
        return self.user_position_lens_mk1.functions.getVaultsForUser(
            str(self._crypto.address)  # type: ignore
        ).call()

    def get_option_chain(self):
        """
        Call the dhv_lens contract to retrieve the options chain.
        """
        return self.dhv_lens_mk1.functions.getOptionChain().call()

    def get_otoken(self, series: Dict[str, Any]):
        """
        Call the option registry to retrieve the otoken.
        underlying: address, underlying asset address is weth lower
        strikeAsset: address,
        expiration: unit256,
        isPut: bool (true if put, false if call)
        strike: unit256 (1e18)
        collateral: address Collateral asset address
        """
        arguments = (
            self._collateral_factory.WETH,
            self._collateral_factory.USDC,
            series["expiration"],
            series["isPut"],
            series["strike"],
            series["collateral"],
        )
        return self.option_registry.functions.getOtoken(*arguments).call()

    def settle_vault(self, vault_id: int):
        """Build the transaction to settle the vault."""
        operate_tuple = [
            {
                "actionType": ActionType.SETTLE_VAULT.value,
                "amount": 0,
                "asset": NULL_ADDRESS,
                "data": NULL_ADDRESS,
                "index": 0,
                "owner": self._crypto.address,  # type: ignore
                "secondAddress": self._crypto.address,  # type: ignore
                "vaultId": vault_id,
            },
        ]
        return self._operate(operate_tuple, self.opyn_controller)

    def redeem_otoken(
        self,
        otoken_id: str,
        amount: int,
    ):
        """
        Build the transaction to redeem the otoken.
        """
        self._logger.info(
            f"Redeeming {amount} of {self.web3.toChecksumAddress(otoken_id)}"
        )
        amount = int(amount * 1e8)
        operate_tuple = [
            {
                "actionType": 8,
                "owner": NULL_ADDRESS,
                "secondAddress": self._crypto.address,  # type: ignore
                "asset": otoken_id,
                "vaultId": 0,
                "amount": amount,
                "index": 0,
                "data": NULL_DATA,
            }
        ]

        return self._operate(operate_tuple, self.opyn_controller)

    def get_otoken_balance(self, otoken_id: str):
        """
        Get the otoken balance.
        """
        otoken = get_contract("o_token", self.web3, chain=self.chain, address=otoken_id)
        return otoken.functions.balanceOf(self._crypto.address).call()  # type: ignore

    def close_long(
        self,
        acceptable_premium: int,
        amount: int,
        otoken_address: str,
    ):
        """
        Build the transaction to close a long position.
        """
        operate_tuple = self._operation_factory.close_long(
            acceptable_premium=acceptable_premium,
            owner_address=self._crypto.address,  # type: ignore
            otoken_address=otoken_address,
            amount=int(amount),
        )
        return self._operate(operate_tuple, self.option_exchange)

    def get_series_info(self, otoken_address: str) -> Dict[str, Any]:
        """
        Returns the series info.
        """
        return self.option_registry.functions.getSeriesInfo(otoken_address).call()

    def close_short(
        self,
        acceptable_premium: int,
        amount: int,
        otoken_address: str,
        collateral_amount: int,
        collateral_asset: str,
        vault_id: int,
    ):
        """
        Build the transaction to close a short position.
        """
        operate_tuple = self._operation_factory.close_short(
            acceptable_premium=acceptable_premium,
            owner_address=self._crypto.address,  # type: ignore
            otoken_address=self.web3.toChecksumAddress(otoken_address),
            amount=int(amount),
            collateral_amount=int(collateral_amount),
            collateral_asset=collateral_asset,
            vault_id=vault_id,
        )
        return self._operate(operate_tuple, self.option_exchange)

    def _operate(
        self,
        operate_tuple: List[Dict[str, Any]],
        contract: Contract,
    ):
        if self._verbose:
            print_operate_tuple(operate_tuple)
        return contract.functions.operate(operate_tuple).build_transaction(
            {
                **{
                    "from": self._crypto.address,  # type: ignore
                    "nonce": self.web3.eth.get_transaction_count(self._crypto.address),  # type: ignore
                },
                **self._default_tx_params,
            }
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 eightballer
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

from typing import Any

from aea.common import JSONLike
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi

from rysk_client.packages.eightballer.contracts.beyond_pricer import PUBLIC_ID


class BeyondPricer(Contract):  # pylint: disable=too-many-public-methods
    """The scaffold contract class for a smart contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        del ledger_api, contract_address, kwargs
        raise NotImplementedError

    @classmethod
    def address_book(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'address_book' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.addressBook().call()
        return {"address": result}

    @classmethod
    def authority(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'authority' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.authority().call()
        return {"address": result}

    @classmethod
    def bid_ask_i_v_spread(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'bid_ask_i_v_spread' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.bidAskIVSpread().call()
        return {"int": result}

    @classmethod
    def collateral_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'collateral_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.collateralAsset().call()
        return {"address": result}

    @classmethod
    def collateral_lending_rate(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'collateral_lending_rate' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.collateralLendingRate().call()
        return {"int": result}

    @classmethod
    def delta_band_width(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'delta_band_width' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.deltaBandWidth().call()
        return {"int": result}

    @classmethod
    def delta_borrow_rates(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'delta_borrow_rates' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.deltaBorrowRates().call()
        return {
            "sellLong": result,
            "sellShort": result,
            "buyLong": result,
            "buyShort": result,
        }

    @classmethod
    def fee_per_contract(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'fee_per_contract' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.feePerContract().call()
        return {"int": result}

    @classmethod
    def get_call_slippage_gradient_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_call_slippage_gradient_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getCallSlippageGradientMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def get_call_spread_collateral_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_call_spread_collateral_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getCallSpreadCollateralMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def get_call_spread_delta_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_call_spread_delta_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getCallSpreadDeltaMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def get_put_slippage_gradient_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_put_slippage_gradient_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getPutSlippageGradientMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def get_put_spread_collateral_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_put_spread_collateral_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getPutSpreadCollateralMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def get_put_spread_delta_multipliers(
        cls, ledger_api: LedgerApi, contract_address: str, _tenor_index: int
    ) -> JSONLike:
        """Handler method for the 'get_put_spread_delta_multipliers' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.getPutSpreadDeltaMultipliers(
            _tenor_index=_tenor_index
        ).call()
        return {"list[int]": result}

    @classmethod
    def liquidity_pool(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'liquidity_pool' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.liquidityPool().call()
        return {"address": result}

    @classmethod
    def low_delta_sell_option_flat_i_v(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'low_delta_sell_option_flat_i_v' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.lowDeltaSellOptionFlatIV().call()
        return {"int": result}

    @classmethod
    def low_delta_threshold(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'low_delta_threshold' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.lowDeltaThreshold().call()
        return {"int": result}

    @classmethod
    def max_tenor_value(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'max_tenor_value' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.maxTenorValue().call()
        return {"int": result}

    @classmethod
    def number_of_tenors(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'number_of_tenors' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.numberOfTenors().call()
        return {"int": result}

    @classmethod
    def protocol(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'protocol' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.protocol().call()
        return {"address": result}

    @classmethod
    def quote_option_price(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        _option_series: tuple,
        _amount: int,
        is_sell: bool,
        net_dhv_exposure: int,
    ) -> JSONLike:
        """Handler method for the 'quote_option_price' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.quoteOptionPrice(
            _optionSeries=_option_series,
            _amount=_amount,
            isSell=is_sell,
            netDhvExposure=net_dhv_exposure,
        ).call()
        return {"totalPremium": result, "totalDelta": result, "totalFees": result}

    @classmethod
    def risk_free_rate(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'risk_free_rate' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.riskFreeRate().call()
        return {"int": result}

    @classmethod
    def slippage_gradient(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'slippage_gradient' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.slippageGradient().call()
        return {"int": result}

    @classmethod
    def strike_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'strike_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.strikeAsset().call()
        return {"address": result}

    @classmethod
    def underlying_asset(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> JSONLike:
        """Handler method for the 'underlying_asset' requests."""
        instance = cls.get_instance(ledger_api, contract_address)
        result = instance.functions.underlyingAsset().call()
        return {"address": result}



"""
Collateral supported by the protocol.
"""
from dataclasses import dataclass

from rysk_client.src.constants import PROTOCOL_DEPLOYMENTS, Chain


@dataclass
class CollateralFactory:
    """Class to return the collateral enum from the deployment."""

    chain: Chain

    @property
    def WETH(self):  # pylint: disable=invalid-name
        """Returns the WETH address"""
        return PROTOCOL_DEPLOYMENTS[self.chain.name].contracts["weth"].address

    @property
    def USDC(self):  # pylint: disable=invalid-name
        """Returns the USDC address"""
        return PROTOCOL_DEPLOYMENTS[self.chain.name].contracts["usdc"].address

    def from_symbol(self, symbol):
        """Returns the address of the collateral from its symbol"""
        if symbol.upper() == "WETH":
            return self.WETH
        if symbol.upper() == "USDC":
            return self.USDC
        raise ValueError(f"Collateral {symbol} not supported")

    def is_supported(self, collateral):
        """Is the sset supported"""
        try:
            self.from_symbol(collateral)
        except ValueError:
            return False
        return True



"""
This file contains all the constants used in the rysk_client package.
"""
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional


def from_camel_case_to_snake_case(string: str):
    """Convert a string from camel case to snake case.
    Note: If the string is all uppercase, it will be converted to lowercase.
    """
    if string.isupper():
        return string.lower()
    return "".join("_" + c.lower() if c.isupper() else c for c in string).lstrip("_")


DEFAULT_TIMEOUT = 10
DEFAULT_ENCODING = "utf-8"
NULL_ADDRESS = "0x0000000000000000000000000000000000000000"
NULL_DATA = "0x0000000000000000000000000000000000000000"
SUPPORTED_LEVERAGES = [1, 1.5, 2, 3]
EMPTY_SERIES = {
    "expiration": 1,
    "strike": 1,
    "isPut": True,
    "collateral": NULL_ADDRESS,
    "underlying": NULL_ADDRESS,
    "strikeAsset": NULL_ADDRESS,
}


CONTRACT_ADDRESSES = {
    "arbitrum": {
        "OpynController": "0x594bD4eC29F7900AE29549c140Ac53b5240d4019",
        "OpynOracle": "0xBA1880CFFE38DD13771CB03De896460baf7dA1E7",
        "OpynNewCalculator": "0x749a3624ad2a001F935E3319743f53Ecc7466358",
        "OpynOptionRegistry": "0x8Bc23878981a207860bA4B185fD065f4fd3c7725",
        "priceFeed": "0x7f86AC0c38bbc3211c610abE3841847fe19590A4",
        "liquidityPool": "0x217749d9017cB87712654422a1F5856AAA147b80",
        "portfolioValuesFeed": "0x7f9d820CFc109686F2ca096fFA93dd497b91C073",
        "optionHandler": "0xc63717c4436043781a63C8c64B02Ff774350e8F8",
        "optionExchange": "0xC117bf3103bd09552F9a721F0B8Bce9843aaE1fa",
        "beyondPricer": "0xeA5Fb118862876f249Ff0b3e7fb25fEb38158def",
        "d_h_v_lens": "0x10779CAE21C91897a5AdD1831Ffb813803c7fcf1",
        "UserPositionLens": "0x02eFd4e61C1883A0FfF1044ACd61c9100859336c",
        "USDC": "0xaf88d065e77c8cC2239327C5EDb3A432268e5831",
        "weth": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
        "o_token": "0x1d96E828e0Aa743783919B24ccDB971504a96C77",
    },
    "arbitrum-goerli": {
        "OpynController": "0x11a602a5F5D823c103bb8b7184e22391Aae5F4C2",
        "OpynOracle": "0x35578F5A49E1f1Cf34ed780B46A0BdABA23D4C0b",
        "OpynNewCalculator": "0xcD270e755C2653e806e16dD3f78E16C89B7a1c9e",
        "OpynOptionRegistry": "0x4E89cc3215AF050Ceb63Ca62470eeC7C1A66F737",
        "priceFeed": "0xf7B1e3a7856067BEcee81FdE0DD38d923b99554D",
        "liquidityPool": "0x0B1Bf5fb77AA36cD48Baa1395Bc2B5fa0f135d8C",
        "portfolioValuesFeed": "0x84fbb7C0a210e5e3A9f7707e1Fb725ADcf0CF528",
        "optionHandler": "0x1F63F3B37f818f05ebefaCa11086e5250958e0D8",
        "optionExchange": "0xb672fE86693bF6f3b034730f5d2C77C8844d6b45",
        "beyondPricer": "0xc939df369C0Fc240C975A6dEEEE77d87bCFaC259",
        "UserPositionLens": "0xa6e2ebD13Cbb085659fB8Ce87fAFdF052066017f",
        "d_h_v_lens": "0xFC2245435e38C6EAd5Cb05ac4ef536A6226eCacb",
        "usdc": "0x408c5755b5c7a0a28D851558eA3636CfC5b5b19d",
        "weth": "0x3b3a1dE07439eeb04492Fa64A889eE25A130CDd3",
        "o_token": "0xB19d2eA6f662b13F530CB84B048877E5Ed0bD8FE",
    },
}


@dataclass
class Contract:
    """Contract dataclass."""

    path: str
    address: str


@dataclass
class Chain:
    """Chain dataclass."""

    name: str
    chain_id: int
    rpc_url: str
    wss_url: str

    def __hash__(self) -> int:
        """Hash the chain."""
        return hash(self.chain_id)


@dataclass
class ProtocolDeployment:
    """Protocol deployment dataclass."""

    name: str
    contracts: Dict[str, Contract]
    chain: Chain
    subgraph_url: Optional[str] = None


WS_URL = "wss://quaint-billowing-morning.arbitrum-goerli.discover.quiknode.pro/def6c4c783fc626cb8a07d38f845b76b458e6e84"
ARBITRUM = Chain(
    name="arbitrum",
    chain_id=42161,
    rpc_url="https://arbitrum.public-rpc.com",
    wss_url="wss://arb1.arbitrum.io/ws",
)
ARBITRUM_GOERLI = Chain(
    name="arbitrum-goerli",
    chain_id=421613,
    rpc_url="https://arbitrum-goerli.rpc.thirdweb.com",
    wss_url=WS_URL,
)

LOCAL_FORK = Chain(
    name="local-fork",
    chain_id=421611,
    rpc_url="http://localhost:8545",
    wss_url="ws://localhost:8545",
)


PROTOCOL_DEPLOYMENTS = {}
SUPPORTED_CHAINS = [ARBITRUM, ARBITRUM_GOERLI]


CHAINS_TO_SUBGRAPH_URL = {
    ARBITRUM: "https://api.goldsky.com/api/public/project_clhf7zaco0n9j490ce421agn4/subgraphs/arbitrum-one/0.1.17/gn",
    ARBITRUM_GOERLI: "https://api.goldsky.com/api/public/project_clhf7zaco0n9j490ce421agn4/subgraphs/devey/0.1.17/gn",
}

for chain in SUPPORTED_CHAINS:
    for name, address in CONTRACT_ADDRESSES[chain.name].items():
        PROTOCOL_DEPLOYMENTS[chain.name] = ProtocolDeployment(
            name=name,
            contracts={
                from_camel_case_to_snake_case(name): Contract(
                    path=Path(os.path.dirname(__file__))
                    / ".."
                    / "packages"
                    / "eightballer"
                    / "contracts"
                    / from_camel_case_to_snake_case(name)
                    / "build"
                    / f"{from_camel_case_to_snake_case(name)}.json",
                    address=address,
                )
                for name, address in CONTRACT_ADDRESSES[chain.name].items()
            },
            chain=chain,
            subgraph_url=CHAINS_TO_SUBGRAPH_URL[chain],
        )


CHAIN_ID_TO_DEPLOYMENT = {
    deploy.chain.chain_id: deploy for deploy in PROTOCOL_DEPLOYMENTS.values()
}

WETH_MULTIPLIER = 1e18
USDC_MULTIPLIER = 1e6



"""
Position Manager
"""
from typing import Dict, List, Optional

from rysk_client.src.position import Position


class PositionManager:
    """Class to manage positions"""

    positions: List[Position]
    collateral: Dict[str, float]

    def __init__(self):
        self.positions = []
        self.collateral = {}
        self.pnl = {}
        self.collateral_value = 0
        self.collateral_value_usd = 0

    def add_position(self, position: Position):
        """Adds a position to the position manager"""

    def update_position(self, position: Position):
        """Updates a position in the position manager"""

    def get_positions(self, address: Optional[str] = None) -> List[Position]:
        """Returns a list of positions"""
        if address:
            return [
                position
                for position in self.positions
                if position.owner_address == address
            ]
        return self.positions

    def _fetch_positions(self):
        """Fetches the positions from the API"""

    def _fetch_orders(self):
        """Fetches the orders from the API"""



"""
Simple client for the rysk contracts implemented in python.
"""
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

import web3

from rysk_client.src.collateral import CollateralFactory
from rysk_client.src.constants import (ARBITRUM_GOERLI, CHAINS_TO_SUBGRAPH_URL,
                                       USDC_MULTIPLIER, WETH_MULTIPLIER, Chain)
from rysk_client.src.crypto import EthCrypto
from rysk_client.src.operation_factory import OperationFactory
from rysk_client.src.order_side import OrderSide
from rysk_client.src.pnl_calculator import PnlCalculator, Trade
from rysk_client.src.position_side import PositionSide
from rysk_client.src.rysk_option_market import (MarketFactory, OptionChain,
                                                RyskOptionMarket)
from rysk_client.src.subgraph import SubgraphClient
from rysk_client.src.utils import get_contract, get_logger
from rysk_client.web3_client import Web3Client, print_operate_tuple

ALLOWED_SLIPPAGE = 0.15

NULL_SERIES = (
    0,
    0,
    False,
    "0x0000000000000000000000000000000000000000",
    "0x0000000000000000000000000000000000000000",
    "0x0000000000000000000000000000000000000000",
)


class ApprovalException(Exception):
    """
    Exception raised when the approval fails.
    """


def from_timestamp(date_string):
    """Parse a timestamp. 2023-05-31T08:00:00.000Z"""
    return datetime.fromtimestamp(int(date_string)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def to_human_format(row):
    """
    Format the row to align to the ccxt unified client.
    'ETH-16MAY23-1550-C'
    """

    month_code = row["expiration_datetime"].strftime("%b").upper()
    day = row["expiration_datetime"].strftime("%d")
    year = str(row["expiration_datetime"].year)[2:]
    strike_price = str(int(int(row["strike"]) / WETH_MULTIPLIER))

    return f"ETH-{day}{month_code}{year}-{strike_price}-{'P' if row['isPut'] else 'C'}"


DEFAULT_MARKET = {
    "base": "ETH",
    "baseId": "ETH",
    "contract": True,
    "contractSize": 1.0,
    "spot": False,
    "swap": False,
    "future": False,
    "type": "option",
    "linear": False,
    "inverse": True,
    "maker": 0.0003,
    "taker": 0.0003,
}


@dataclass
class RyskClient:  # noqa: R0902
    """
    Client for the rysk contracts.
    """

    _markets: List[RyskOptionMarket]
    _tickers: List[Dict[str, Any]]
    _otokens: Dict[str, Dict[str, Any]]
    web3_client: Web3Client

    def __init__(
        self,
        address: Optional[str] = None,
        private_key: Optional[str] = None,
        logger=None,
        chain: Chain = ARBITRUM_GOERLI,
        verbose: bool = True,
    ):
        self._markets: List[RyskOptionMarket] = []
        self._tickers = []
        self._otokens = {}
        self._option_chain: OptionChain
        self._crypto = EthCrypto(address, private_key)
        self._logger = logger or get_logger()
        self.web3_client = Web3Client(
            self._logger,
            self._crypto,
            chain=chain,
            verbose=verbose,
        )
        self.subgraph_client = SubgraphClient(CHAINS_TO_SUBGRAPH_URL[chain])
        self._logger.info(
            f"Rysk client initialized and connected to the blockchain at {self.web3_client.web3.provider}"
        )
        self._verbose = verbose
        self.operation_factory = OperationFactory(chain)
        self.collateral_factory = CollateralFactory(chain)
        self.market_factory = MarketFactory(chain)

    def _sign_and_submit(self, txn, retries=3, backoff=2):
        """
        Sign and submit transaction.

        retries: number of retries
        backoff: backoff in seconds

        """
        try:
            txn["nonce"] = self.web3_client.web3.eth.get_transaction_count(
                self._crypto.address
            )
            signed_txn = self.web3_client.web3.eth.account.sign_transaction(
                txn, private_key=self._crypto.private_key
            )
            tx_hash = self.web3_client.web3.eth.send_raw_transaction(
                signed_txn.rawTransaction
            )
            self._logger.debug(f"Transaction hash: {tx_hash.hex()}")
            receipt = self.web3_client.web3.eth.wait_for_transaction_receipt(
                tx_hash, 180
            )
            # we need to check the receipt to see if the transaction was successful
            if receipt["status"] == 0:
                self._logger.error(f"Transaction failed: {receipt}")
                raise ValueError("Transaction Failed!")
            self._logger.debug(f"Transaction successful: {receipt}")
            return tx_hash.hex()

        except ValueError as error:
            if retries > 0:
                self._logger.warning(
                    f"Error {error} while submitting transaction. Retrying..."
                )
                time.sleep(backoff)
                return self._sign_and_submit(
                    txn, retries=retries - 1, backoff=backoff * 2
                )
            raise error

    def fetch_markets(self) -> List[Dict[str, Any]]:
        """
        Fetchs the markets from the subgraph.
        """

        raw_data = self.web3_client.get_option_chain()
        self._option_chain = OptionChain(raw_data)
        self._markets = [
            RyskOptionMarket.from_option_drill(*data)
            for data in self._option_chain.active_markets
        ]
        return [market.to_json() for market in self._markets]  # type: ignore

    def to_checksum_address(self, address):
        """Convert an address to a checksum address."""
        return self.web3_client.web3.toChecksumAddress(address)

    def fetch_tickers(
        self, market: Optional[str] = None, is_active: Optional[bool] = True
    ) -> List[Dict[str, Any]]:
        """
        Fetchs the ticker from the beyond pricer smart contract.
        """

        if not self._markets:
            self.fetch_markets()

        tradeable = filter(lambda x: x["active"] is is_active, self._markets)  # type: ignore

        if market:
            tradeable = filter(lambda x: x["id"] == market, tradeable)  # type: ignore

        return [market.to_json() for market in self._markets]  # type: ignore

    @property
    def otokens(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns a dictionary of the otokens.
        """
        if not self._otokens:
            self.fetch_tickers()
        self._otokens = {ticker["info"]["id"]: ticker for ticker in self._tickers}
        return self._otokens

    def fetch_positions(self, expired=False) -> List[Dict[str, Any]]:
        """
        Fetchs the positions from the subgraph.
        """

        if self._crypto.address is None:
            raise ValueError("No account address was provided.")

        longs = self.subgraph_client.query_longs(address=self._crypto.address)
        shorts = self.subgraph_client.query_shorts(address=self._crypto.address)

        parsed_short = [self._parse_position(pos, PositionSide.SHORT) for pos in shorts]
        parsed_longs = [self._parse_position(pos, PositionSide.LONG) for pos in longs]

        if expired:
            positions = filter(
                lambda x: x["datetime"] <= datetime.now(),
                parsed_longs + parsed_short,
            )
        else:
            positions = filter(
                lambda x: x["datetime"] > datetime.now(),
                parsed_longs + parsed_short,
            )
        return list(positions)

    def _parse_position(
        self, position: Dict[str, Any], side: PositionSide
    ) -> Dict[str, Any]:
        """
        Parse the position data from the subgraph into a unified format.
        """
        position["expiration_datetime"] = datetime.fromtimestamp(
            int(position["oToken"]["expiryTimestamp"])
        )
        position["strike"] = float(position["oToken"]["strikePrice"]) * 1e10
        position["isPut"] = position["oToken"]["isPut"]

        pnl_calculator = PnlCalculator()

        buys = [
            Trade(
                int(order["amount"]) / WETH_MULTIPLIER,
                total_cost=int(order["premium"]) / USDC_MULTIPLIER,
            )
            for order in position["optionsBoughtTransactions"]
        ]
        sells = [
            Trade(
                -int(order["amount"]) / WETH_MULTIPLIER,
                total_cost=-int(order["premium"]) / USDC_MULTIPLIER,
            )
            for order in position["optionsSoldTransactions"]
        ]

        pnl_calculator.add_trades(buys + sells)

        symbol = to_human_format(position)

        if symbol in self.otokens:
            self._logger.debug(f"Found `{symbol}` in the otokens list.")
            book_side = "ask" if side == PositionSide.LONG else "bid"
            price = self.otokens[symbol][book_side]
            pnl_calculator.update_price(price)
        else:
            self._logger.debug(
                f"Could not find `{symbol}` in the otokens list. Maket is probably not active."
            )
        result = {
            "id": position["id"],
            "symbol": symbol,
            "timestamp": int(position["oToken"]["expiryTimestamp"]) * 1000,
            "datetime": datetime.fromtimestamp(
                int(position["oToken"]["expiryTimestamp"])
            ),
            "initialMarginPercentage": None,
            "realizedPnl": pnl_calculator.realised_pnl,
            "unrealizedPnl": pnl_calculator.unrealised_pnl,
            "contractSize": pnl_calculator.position_size,
            "side": side.value,
            "size": pnl_calculator.position_size,
            "info": position,
            "entryPrice": pnl_calculator.average_price,
        }
        return result

    def create_order(
        self,
        symbol: str,
        amount: float,
        side: str = "buy",
    ) -> Dict[str, Any]:
        """Create a market order."""
        if side.upper() not in OrderSide.__members__:
            raise ValueError("Invalid order side")

        if side == OrderSide.BUY.value:
            transaction = self.buy_option(symbol, amount)
        else:
            transaction = self.sell_option(symbol, amount)
        submitted = self._sign_and_submit(transaction)
        self.web3_client.web3.eth.wait_for_transaction_receipt(submitted)
        self._logger.info(f"Submitted transaction with hash: {submitted}")
        return {
            "id": submitted,
            "symbol": symbol,
            "datetime": datetime.now(),
        }

    def get_market(self, symbol: str) -> RyskOptionMarket:
        """
        Returns the market information.
        """
        if not self._markets:
            self.fetch_markets()

        for market in self._markets:
            if market.name == symbol:
                return market
        raise ValueError(f"Could not find market {symbol}.")

    def buy_option(  # noqa: R0914
        self,
        market: str,
        amount: float,
        collateral_asset: str = "USDC",
        leverage: float = 1,
    ):
        """
        Create a buy option order.
        """
        # now we can try to buy the option
        if not self._markets:
            self._logger.info("Fetching Tickers.")
            self.fetch_tickers()  # type: ignore

        self._logger.info(f"Fetching acceptable premium for {market}")
        rysk_option_market = self.get_market(market)
        rysk_option_market.collateral = self.collateral_factory.from_symbol(
            collateral_asset
        )
        self._logger.info(
            f"Buying {amount} of {market} with {collateral_asset} collateral @ {leverage}x leverage."
        )
        # we first check the approval amount of the collateral asset
        if not self.collateral_factory.is_supported(collateral_asset):
            raise ValueError(
                f"Collateral asset {collateral_asset} is not supported by the protocol."
            )
        collateral_contract = get_contract(
            collateral_asset.lower(), self.web3_client.web3, self.web3_client.chain
        )
        collateral_approved = self.web3_client.is_approved(
            collateral_contract,
            self.web3_client.option_exchange.address,
            str(self._crypto.address),
            int(amount * WETH_MULTIPLIER),
        )
        self._logger.info(f"Collateral approved: {collateral_approved}.")

        if not collateral_approved:
            self._logger.info(
                f"Approving {collateral_asset} collateral for {self.web3_client.option_exchange.address}"
            )

            txn = self.web3_client.create_approval(
                collateral_contract,
                self.web3_client.option_exchange.address,
                str(self._crypto.address),
                int(amount * WETH_MULTIPLIER),
            )
            # we submit and sign the transaction
            result = self._sign_and_submit(txn)
            self._logger.info(f"Transaction successful with hash: {result}")

        series = self.market_factory.to_series(rysk_option_market)

        otoken_address = self.web3_client.get_otoken(series)
        balance = self.web3_client.get_otoken_balance(otoken_address)
        self._logger.info(f"Balance of {otoken_address}: {balance}")

        self._logger.info(
            f"Fetching market data for {market}. Otoken address: {otoken_address}"
        )
        _amount = amount * WETH_MULTIPLIER
        acceptable_premium = self.web3_client.get_options_prices(  # type: ignore
            series,
            rysk_option_market.dhv,
            side=OrderSide.BUY.value,
            amount=_amount,
        )  # pylint: disable=E1120

        self._logger.info(
            f"Acceptable premium: ${acceptable_premium / USDC_MULTIPLIER:.2f}"
        )

        # we check if we need to issue the option
        issuance_required = self.is_issuance_required(otoken_address)

        operate_tuple = self.operation_factory.buy(
            int(acceptable_premium * (1 + ALLOWED_SLIPPAGE)),
            owner_address=self._crypto.address,  # pylint: disable=E1120
            amount=int(_amount),
            option_market=rysk_option_market,
            issuance_required=issuance_required,
        )

        if self._verbose:
            print_operate_tuple([operate_tuple])

        try:
            txn = self.web3_client.option_exchange.functions.operate(
                operate_tuple
            ).build_transaction({"from": self._crypto.address})
        except web3.exceptions.ContractCustomError as error:  # pylint: disable=E1101
            self._logger.error("Transaction failed due to incorrect parameters.")
            raise ValueError(error) from error

        return txn

    def is_issuance_required(self, otoken_address: str) -> bool:
        """
        Returns True if an issuance is required.
        """
        result = self.web3_client.get_series_info(otoken_address)
        return result == NULL_SERIES

    def sell_option(  # noqa
        self,
        market: str,
        amount: float,
        collateral_asset: str = "weth",
        leverage: float = 1,
    ):
        """
        Create a sell option order.
        """
        self._logger.info(
            f"Selling {amount} of {market} with {collateral_asset} collateral @ {leverage}x leverage."
        )

        if not self._markets:
            self._logger.info("Fetching Tickers.")
            self.fetch_tickers()
        _amount = amount * WETH_MULTIPLIER

        rysk_option_market = self.get_market(market)
        if rysk_option_market.is_put:
            rysk_option_market.collateral = self.collateral_factory.USDC
            # we need to approve the collateral
            amount_to_approve = int(
                rysk_option_market.strike / WETH_MULTIPLIER * amount * USDC_MULTIPLIER
            )
            contract = self.web3_client.usdc
            collateral_asset = "usdc"

        else:
            rysk_option_market.collateral = self.collateral_factory.WETH
            amount_to_approve = int(_amount * (1 + ALLOWED_SLIPPAGE))
            contract = self.web3_client.settlement_weth
        series = self.market_factory.to_series(rysk_option_market)

        acceptable_premium = self.web3_client.get_options_prices(
            series,
            dhv_exposure=rysk_option_market.dhv,
            amount=_amount,
            side=OrderSide.SELL.value,
            collateral=collateral_asset,
        )

        self._logger.info(
            f"Acceptable premium: ${acceptable_premium / USDC_MULTIPLIER:.2f}"
        )

        user_vaults = self.web3_client.fetch_user_vaults(self._crypto.address)
        otoken_id = self.web3_client.get_otoken(series)
        self._logger.info(f"Option Otoken id is {otoken_id}")

        issue_new_vault = False
        if otoken_id not in set(i[1] for i in user_vaults):
            new_vault_id = len(user_vaults) + 1
            self._logger.info(
                f"Necessary to create a vault for the user. New vault id is {new_vault_id}"
            )
            vault_id = new_vault_id
            issue_new_vault = True

        else:
            # we need to use the vault
            vault_id = [f[1] for f in user_vaults].index(otoken_id) + 1
            self._logger.info(f"Using existing vault id user_vaults {vault_id}")

        # we check the approval of the amount
        self._logger.info(
            f"Checking approval of {amount_to_approve / WETH_MULTIPLIER} of {collateral_asset}"
        )
        allowance = contract.functions.allowance(
            self._crypto.address,
            self.web3_client.option_exchange.address,
        ).call()
        self._logger.info(f"Allowance is {allowance}")
        if allowance < amount_to_approve:
            self._logger.info("Need to approve more collateral.")
            approve_tx = self.web3_client.create_approval(
                contract=contract,
                spender=self.web3_client.option_exchange.address,
                owner=self._crypto.address,
                amount=amount_to_approve,
            )

            self._logger.debug(f"Approve tx is {approve_tx}")
            tx_hash = self._sign_and_submit(approve_tx)
            self._logger.info(f"Tx hash is {tx_hash}")

        # has allowcance incremented
        allowance = contract.functions.allowance(
            self._crypto.address,
            self.web3_client.option_exchange.address,
        ).call()
        self._logger.info(f"Allowance is {allowance}")
        otoken_address = self.web3_client.get_otoken(series)

        operate_tuple = self.operation_factory.sell(
            int(acceptable_premium * 0.95),
            owner_address=self._crypto.address,  # pylint: disable=E1120
            exchange_address=self.web3_client.option_exchange.address,
            otoken_address=otoken_address,
            amount=int(_amount),
            vault_id=int(vault_id),
            collateral_amount=int(amount_to_approve),
            rysk_option_market=rysk_option_market,
            issue_new_vault=issue_new_vault,
        )
        return self.web3_client._operate(  # pylint: disable=protected-access
            operate_tuple, self.web3_client.option_exchange
        )

    def watch_trades(self):
        """Watch trades."""
        self._logger.info("Watching trades...")
        self.web3_client.watch_trades()

    def fetch_balances(self):
        """Fetch balances."""
        self._logger.info("Fetching balances...")
        return self.web3_client.get_balances()

    def settle_vault(self, vault_id):
        """Settle options."""
        self._logger.info(f"Settling vault {vault_id}...")
        txn = self.web3_client.settle_vault(vault_id=vault_id)
        return self._sign_and_submit(txn)

    def redeem_otoken(self, otoken_id: str, amount: int):
        """Redeem otoken."""
        self._logger.info(f"Redeeming otoken {otoken_id}...")
        txn = self.web3_client.redeem_otoken(otoken_id=otoken_id, amount=amount)
        return self._sign_and_submit(txn)

    def redeem_market(self, market: str):
        """Redeem otoken."""
        self._logger.info(f"Redeeming market {market}...")
        rysk_option_market = RyskOptionMarket.from_str(market)
        series = self.market_factory.to_series(rysk_option_market)
        otoken_address = self.web3_client.get_otoken(series)
        amount = self.web3_client.get_otoken_balance(otoken_address) / 10**8
        return self.redeem_otoken(otoken_address, amount)

    @property
    def active_markets(self):
        """Get active markets."""
        if not self._markets:
            self.fetch_markets()
        return {market.name: market for market in self._markets}

    def close_long(self, market: str, size: float):
        """Close long."""
        self._logger.info(f"Closing long {market}...")
        # as we are long, we use usdc as collateral
        collateral_asset = self.collateral_factory.USDC
        _market = self.get_market(market)
        _market.collateral = collateral_asset
        series = self.market_factory.to_series(_market)
        otoken_address = self.web3_client.get_otoken(series)
        if size is None:
            _amount = self.web3_client.get_otoken_balance(otoken_address) * 10**10
        else:
            _amount = size * WETH_MULTIPLIER
        if _market.name not in self.active_markets:
            raise ValueError(f"{market} is not an active market...")
        acceptable_premium = int(_market.ask * (1 - ALLOWED_SLIPPAGE) * size)
        # we check the approval
        otoken_contract = self.web3_client.get_otoken_contract(otoken_address)

        # we get the balance
        balance = self.web3_client.get_otoken_balance(otoken_address)

        self._logger.info(f"Balance for {market} is {balance / 10**8}")

        if balance == 0:
            raise ValueError(f"Nothing to close for {market}...")

        if not self.web3_client.is_approved(
            otoken_contract,
            self.web3_client.option_exchange.address,
            self._crypto.address,
            int(10**8 * _amount),
        ):
            self._logger.info(f"Approving {market}...")
            txn = self.web3_client.create_approval(
                otoken_contract,
                self.web3_client.option_exchange.address,
                self._crypto.address,  # type: ignore
                int(10**8 * _amount),
            )
            self._sign_and_submit(
                txn,
            )

        if _amount == 0:
            raise ValueError(f"Nothing to close for {market}...")

        txn = self.web3_client.close_long(
            acceptable_premium=acceptable_premium,
            amount=_amount,
            otoken_address=self.web3_client.web3.toChecksumAddress(otoken_address),
        )
        return self._sign_and_submit(txn)

    def close_short(self, market: str, size: float):
        """
        CLose short.
        """
        self._logger.info(f"Closing short {market}...")
        # as we are short, we ensure we are covered
        rysk_option_market = self.active_markets[market]
        if rysk_option_market.is_put:
            rysk_option_market.collateral = self.collateral_factory.USDC
        else:
            rysk_option_market.collateral = self.collateral_factory.WETH
        series = self.market_factory.to_series(rysk_option_market)
        otoken_address = self.web3_client.get_otoken(series)

        if size is None:
            raise NotImplementedError(
                "Closing short with no size is not implemented yet"
            )
        _amount = size * WETH_MULTIPLIER
        if rysk_option_market.name not in self.active_markets:
            raise ValueError(f"{market} is not an active market...")
        acceptable_premium = int(rysk_option_market.bid * (1 + ALLOWED_SLIPPAGE) * size)

        user_vaults = self.web3_client.fetch_user_vaults(self._crypto.address)
        otoken_id = self.web3_client.get_otoken(series)
        self._logger.info(f"Option Otoken id is {otoken_id}")

        positions = [f for f in user_vaults if f[1] == otoken_id]
        if len(positions) == 0:
            raise ValueError(f"Nothing to close for {market}...")
        if len(positions) > 1:
            raise ValueError(f"Multiple positions for {market}...")
        vault_id = positions[0][0]

        if rysk_option_market.is_put:
            # here we retrieve how much collateral we get for the amount of options
            # we basically need strike * amount
            strike = rysk_option_market.strike / WETH_MULTIPLIER
            collateral_amount = int(strike * size * USDC_MULTIPLIER)
        else:
            collateral_amount = int(_amount)

        txn = self.web3_client.close_short(
            acceptable_premium=acceptable_premium,
            amount=int(_amount),
            otoken_address=self.web3_client.web3.toChecksumAddress(otoken_address),
            collateral_asset=rysk_option_market.collateral,
            collateral_amount=collateral_amount,
            vault_id=vault_id,
        )
        return self._sign_and_submit(txn)



"""
Crypto module.
"""
from dataclasses import dataclass
from typing import Optional


@dataclass
class EthCrypto:
    """Represents a crypto wallet."""

    address: Optional[str]
    private_key: Optional[str]
